/*
 * Copyright (c) 2010, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2010-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net;

import java.io.IOException;
import java.net.Inet4Address;
import java.net.Inet6Address;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;
import java.net.PortUnreachableException;
import java.net.SocketAddress;
import java.net.SocketException;
import java.util.Enumeration;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.natives.DatagramSocketChannelNatives;
import com.is2t.support.net.natives.MulticastSocketChannelNatives;
import com.is2t.support.net.util.IPAddressConstants;
import com.is2t.support.net.util.IntRef;
import com.is2t.support.net.util.NativeIOExceptionHandler;

import ej.bon.ByteArray;
import ej.bon.Util;
import ej.error.Message;
import ej.sni.NativeIOException;

public class DatagramSocketChannel extends SocketChannel {

	/**
	 * The maximum host data size 16 bytes max for the host address (required for IPv6) + 4 bytes for the host port
	 */
	private static final int HOST_DATA_SIZE = 20;

	public DatagramSocketChannel() throws IOException {
		super();
	}

	public void initDatagramSocket(boolean isMulticast) throws IOException {
		if (isMulticast) {
			initMulticastSocket();
		} else {
			initSocket(false);
		}
	}

	@Override
	public boolean isStream() {
		return false;
	}

	public void send(byte[] data, int offset, int length, InetSocketAddress dst) throws IOException {
		InetAddress addr = dst.getAddress();
		if (addr == null) {
			throw new NullPointerException();
		}

		byte[] hostAddr = addr.getAddress();
		int dstAddrLength = addr.getAddress().length;
		if ((dstAddrLength == IPAddressConstants.IPV4_ADDRESS_SIZE)
				|| (dstAddrLength == IPAddressConstants.IPV6_ADDRESS_SIZE)) {
			int fd = this.nfd.getNativeFD();
			int dstPort = dst.getPort();

			try {
				DatagramSocketChannelNatives.send(fd, data, offset, length, hostAddr, hostAddr.length, dstPort);
			} catch (NativeIOException e) {
				throw handleNativeIOException(e);
			}
		} else {
			throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
		}
	}

	public SocketAddress receive(byte[] data, int offset, int length, IntRef readLength, int timeout)
			throws IOException {

		byte[] hostPort = new byte[HOST_DATA_SIZE];
		try {
			int fd = this.nfd.getNativeFD();
			int hostlen = 0;
			int receiveLength = 0;
			long absoluteTimeout = 0;
			if (timeout > 0) {
				absoluteTimeout = Util.platformTimeMillis() + timeout;
			}

			// The receive operation blocks until the fd become readable
			long ret = DatagramSocketChannelNatives.receive(fd, data, offset, length, hostPort, hostPort.length,
					absoluteTimeout);

			// extract host length and data receive lenght from the long value returned by
			// native
			hostlen = (int) (ret & 0xFFFFFFFFl);
			receiveLength = (int) ((ret >> 32l) & 0xFFFFFFFFl);

			// update the read value
			readLength.setValue(receiveLength);

			if (hostlen <= 0) {
				return null;
			} else {
				byte[] addr = new byte[hostlen];
				int addrLength = addr.length;
				System.arraycopy(hostPort, 0, addr, 0, addrLength);
				int port = ByteArray.readInt(hostPort, addrLength);

				InetAddress inetAddress;
				if (hostlen == IPAddressConstants.IPV4_ADDRESS_SIZE) { // IPv4
					inetAddress = Inet4Address.getByAddress(addr);
				} else if (hostlen == IPAddressConstants.IPV6_ADDRESS_SIZE) { // IPv6
					inetAddress = Inet6Address.getByAddress(addr);
				} else {
					throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
				}

				return new InetSocketAddress(inetAddress, port);
			}
		} catch (NativeIOException e) {
			throw handleNativeIOException(e);
		}
	}

	/**
	 * Disconnect this channel, if it is a datagram socket. Disconnecting a datagram channel will disassociate it from
	 * any address, so the socket will remain open, but can send and receive datagrams from any address.
	 *
	 * @throws IOException
	 *             If disconnecting this channel fails, or if this channel is not a datagram channel.
	 */
	public void disconnect() throws IOException {
		try {
			DatagramSocketChannelNatives.disconnect(this.nfd.getNativeFD());
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeIOException(e);
		}
	}

	public void setTimeToLive(int ttl) throws IOException {
		try {
			MulticastSocketChannelNatives.setTimeToLive(this.nfd.getNativeFD(), ttl);
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeIOException(e);
		}
	}

	public int getTimeToLive() throws IOException {
		try {
			return MulticastSocketChannelNatives.getTimeToLive(this.nfd.getNativeFD());
		} catch (NativeIOException e) {
			throw NativeIOExceptionHandler.generateNativeIOException(e);
		}
	}

	private int fillInterfaceAddress(byte[] address, NetworkInterface netIf) {
		InetAddress addr = null;
		if (netIf != null) {
			Enumeration<InetAddress> addresses = netIf.getInetAddresses();
			if (addresses.hasMoreElements()) {
				addr = addresses.nextElement();
			}
		}
		if (addr != null) {
			byte[] addrBytes = addr.getAddress();
			int len = addrBytes.length;
			System.arraycopy(addrBytes, 0, address, 0, len);
			return len;
		}
		return 0;
	}

	private void joinOrLeave(boolean join, InetAddress addr, NetworkInterface netIf) throws IOException {
		if (addr == null) {
			throw new NullPointerException();
		}

		byte[] mcastAddr = addr.getAddress();
		byte[] netIfAddr = new byte[IPAddressConstants.IPV6_ADDRESS_SIZE];
		int netIfAddrLength = fillInterfaceAddress(netIfAddr, netIf);
		try {
			int addrLength = addr.getAddress().length;
			if ((addrLength == IPAddressConstants.IPV4_ADDRESS_SIZE)
					|| (addrLength == IPAddressConstants.IPV6_ADDRESS_SIZE)) {
				int fd = this.nfd.getNativeFD();
				MulticastSocketChannelNatives.joinOrLeave(fd, join, mcastAddr, mcastAddr.length, netIfAddr,
						netIfAddrLength);
			} else {
				throw new SocketException(Message.at(NetErrors.getInstance(), NetErrors.J_EAFNOSUPPORT));
			}
		} catch (NativeIOException e) {
			throw handleNativeIOException(e);
		}
	}

	public void join(InetAddress addr, NetworkInterface netIf) throws IOException {
		joinOrLeave(true, addr, netIf);
	}

	public void leave(InetAddress addr, NetworkInterface netIf) throws IOException {
		joinOrLeave(false, addr, netIf);
	}

	protected static IOException handleNativeIOException(NativeIOException cause) {
		int errCode = cause.getErrorCode();
		if (errCode == NetErrors.J_ECONNREFUSED) {
			PortUnreachableException ex = new PortUnreachableException(Message.at(NetErrors.getInstance(), errCode));
			ex.initCause(cause);
			return ex;
		} else {
			return NativeIOExceptionHandler.generateNativeIOException(cause);
		}
	}
}
