/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2014-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;
import com.is2t.support.net.util.NativeIOExceptionHandler;

import ej.sni.NativeIOException;

public class ChannelNatives {

	private static boolean initialized = false;

	static {
		initializeNativePart();
	}

	/**
	 * Initialize the NET foundation library native part.
	 */
	public static void initializeNativePart() {
		if (!initialized) {
			try {
				initialize();
			} catch (NativeIOException e) {
				throw NativeIOExceptionHandler.generateNativeRuntimeException(e);
			}
		}
		initialized = true;
	}

	/**
	 * Initializes the TCP/IP stack components.
	 * 
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void initialize() throws NativeIOException;

	/**
	 * Returns option value for the socket associated with the file descriptor {@code fd}.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param option
	 *            the socket option identifier
	 * @return the option value for the socket associated with the file descriptor {@code fd} or a negative error code
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getOption(int fd, int option) throws NativeIOException;

	/**
	 * Sets option value for the socket associated with the file descriptor {@code fd}.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param option
	 *            the socket option identifier
	 * @param value
	 *            the option value to set
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void setOption(int fd, int option, int value) throws NativeIOException;

	/**
	 * Returns option value for the socket associated with the file descriptor {@code fd}.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param option
	 *            the socket option identifier
	 * @param dst
	 *            the destination byte array for the option value
	 * @param dstLength
	 *            the maximum length of the value the array can hold
	 * @return the length of the option value for the socket associated with the file descriptor {@code fd} or a
	 *         negative error code
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getOption(int fd, int option, byte[] dst, int dstLength) throws NativeIOException;

	/**
	 * Sets option value for the socket associated with the file descriptor {@code fd}.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param option
	 *            the socket option identifier
	 * @param value
	 *            the option value to set as a byte array
	 * @param valueLength
	 *            the length of the value in the array
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void setOption(int fd, int option, byte[] value, int valueLength) throws NativeIOException;

	/**
	 * Binds the socket associated with the file descriptor {@code fd} to the specified local address {@code addr} and
	 * {@code port}.
	 *
	 * @param fd
	 *            the native file descriptor
	 * @param addr
	 *            the IP address array, in network byte order
	 * @param length
	 *            the address size (4 bytes long for IPv4 addresses and 16 bytes long for IPv6 addresses)
	 * @param port
	 *            the port to bind to
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void bind(int fd, byte[] addr, int length, int port) throws NativeIOException;

	/**
	 * Closes the socket associated with the file descriptor {@code fd}.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void closeNative(int fd) throws NativeIOException;

	/**
	 * Shuts down the socket associated with the file descriptor {@code fd}'s input/output streams.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void shutdown(int fd) throws NativeIOException;

	/**
	 * Listens on the socket associated with the given file descriptor {@code fd} for incoming connections.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param backlog
	 *            the listen backlog size
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void listen(int fd, int backlog) throws NativeIOException;
}
