/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2014-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;

import ej.sni.NativeIOException;

public class SocketChannelNatives {

	/**
	 * Connects the socket associated with the file descriptor {@code fd} to the remote host address {@code addr} and
	 * {@code port} with a specified absolute timeout {@code absoluteTimeout} value.
	 * 
	 * The connection will then block until established, the given absolute timeout is reached, or an error occurs. If
	 * the timeout expires, an <B>ej.sni.NativeIOException</B> is raised with the error code
	 * <B>NetErrors.J_ETIMEDOUT</B>
	 * 
	 * {@code absoluteTimeout} is an absolute time in milliseconds computed from the system time returned by
	 * <B>ej.bon.Util.platformTimeMillis()</B>. A timeout of zero is interpreted as an infinite timeout.
	 * 
	 * @param fd
	 *            the socket file descriptor
	 * @param addr
	 *            the remote host address buffer
	 * @param length
	 *            the remote host address size
	 * @param port
	 *            the remote host port
	 * @param absoluteTimeout
	 *            The absolute timeout in milliseconds or 0 if no timeout
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native void connect(int fd, byte[] addr, int length, int port, long absoluteTimeout)
			throws NativeIOException;

	/**
	 * Returns the local port to which the socket associated with the file descriptor {@code fd} is bound.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @return the local port to which the socket associated with the file descriptor {@code fd} is bound; -1 if the
	 *         socket is not bound.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getLocalPort(int fd) throws NativeIOException;

	/**
	 * Returns the remote host port to which the socket associated with the file descriptor {@code fd} is connected.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @return the remote port to which the socket associated with the file descriptor {@code fd} is connected.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getPeerPort(int fd) throws NativeIOException;

	/**
	 * Gets the local address to which the socket associated with the file descriptor {@code fd} is bound. When this
	 * methods returns, the {@code addr} buffer is filled with the local address.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param addr
	 *            the buffer into which the local address will be stored
	 * @param length
	 *            the buffer length
	 * @return the local address size in bytes or 0 if the socket is not bound.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getLocalAddress(int fd, byte[] addr, int length) throws NativeIOException;

	/**
	 * Gets the remote host address to which the socket associated with the file descriptor {@code fd} is connected.
	 * When this methods returns, the {@code addr} buffer is filled with the remote host address.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param addr
	 *            the buffer into which the remote host address will be stored
	 * @param length
	 *            the buffer length
	 * @return the remote host address size in bytes or 0 if the socket is not bound.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int getPeerAddress(int fd, byte[] addr, int length) throws NativeIOException;

	/**
	 * Creates a new socket and returns its associated file descriptor.
	 *
	 * @param stream
	 *            set to true for streaming sockets; false for datagrams.
	 * @return the file descriptor of the new socket.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int socket(boolean stream) throws NativeIOException;

	/**
	 * Creates a new server socket, returning the native file descriptor.
	 *
	 * @return the file descriptor of the new server socket.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int serverSocket() throws NativeIOException;

	/**
	 * Creates a new multicast socket, returning the native file descriptor.
	 *
	 * @return the file descriptor of the new server socket.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int multicastSocket() throws NativeIOException;
}
