/*
 * Copyright (c) 2014, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2014-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.natives;

import com.is2t.errors.NetErrors;

import ej.sni.NativeIOException;

public class StreamSocketChannelNatives {

	/**
	 * Reads {@code length} bytes from the socket associated with the file descriptor {@code fd} into the specified
	 * destination buffer {@code dst}. The return number of bytes read could be smaller than the {@code length} bytes
	 * requested.
	 * 
	 * The read operation will block until data is arrived, the given absolute timeout is reached, or an error occurs.
	 * If the timeout expires, an <B>ej.sni.NativeIOException</B> is raised with the error code
	 * <B>NetErrors.J_ETIMEDOUT</B>
	 * 
	 * {@code absoluteTimeout} is an absolute time in milliseconds computed from the system time returned by
	 * <B>ej.bon.Util.platformTimeMillis()</B>. A timeout of zero is interpreted as an infinite timeout.
	 * 
	 * @param fd
	 *            the socket file descriptor
	 * @param dst
	 *            the destination buffer into which the read data is stored
	 * @param offset
	 *            the start offset in the buffer at which the data is written
	 * @param length
	 *            the maximum number of data to read
	 * @param absoluteTimeout
	 *            The absolute timeout in milliseconds or 0 if no timeout
	 * @return the number of bytes read; -1 if there is no more data because the end of the stream has been reached.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int read(int fd, byte[] dst, int offset, int length, long absoluteTimeout)
			throws NativeIOException;

	/**
	 * Writes {@code length} bytes from the specified buffer {@code buf} starting at the {@code offset} to the output
	 * stream of the socket associated with the file descriptor {@code fd}. The write operation will block until all
	 * data have been sent or an error occurs.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @param buf
	 *            the buffer of data to write
	 * @param offset
	 *            the offset in the data buffer
	 * @param length
	 *            the number of data to write
	 * @return the number of bytes written.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int write(int fd, byte[] buf, int offset, int length) throws NativeIOException;

	/**
	 * Accepts an incoming connection for the listening server socket associated with the file descriptor {@code fd} and
	 * returns the file descriptor of the accepted socket.
	 *
	 * The accept operation will block until a new connection is attempted, the given absolute timeout reaches, or an
	 * error occurs. If the timeout expires, an <B>ej.sni.NativeIOException</B> is raised with the error code
	 * <B>NetErrors.J_ETIMEDOUT</B> A timeout of zero is interpreted as an infinite timeout.
	 * 
	 * {@code absoluteTimeout} is an absolute time in milliseconds computed from the system time returned by
	 * <B>ej.bon.Util.platformTimeMillis()</B>. A timeout of zero is interpreted as an infinite timeout.
	 * 
	 * @param fd
	 *            the server socket file descriptor that listening for incoming connection
	 * @param absoluteTimeout
	 *            The absolute timeout in milliseconds or 0 if no timeout
	 * @return the file descriptor of the accepted socket.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int accept(int fd, long absoluteTimeout) throws NativeIOException;

	/**
	 * Returns an estimate of the number of bytes that can be read from the socket associated with the file descriptor
	 * {@code fd} without blocking by the next invocation of a read operation.
	 *
	 * @param fd
	 *            the socket file descriptor
	 * @return an estimate of the number of bytes that can be read from the socket associated with the file descriptor
	 *         {@code fd} without blocking by the next invocation of a read operation.
	 * @throws NativeIOException
	 *             if an error occurs.
	 * @see {@link NetErrors} for error codes
	 */
	public static native int available(int fd) throws NativeIOException;

}
