/*
 * Copyright 2018-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.net.util;

import java.io.IOException;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;

import com.is2t.errors.NetErrors;

import ej.error.Message;
import ej.sni.NativeIOException;

public class NativeIOExceptionHandler {

	public static SocketException generateNativeSocketException(NativeIOException cause) {
		SocketException ex = new SocketException(generateNativeIOExceptionMessage(cause));
		ex.initCause(cause);
		return ex;
	}

	public static UnknownHostException generateNativeUnknownHostException(NativeIOException cause) {
		UnknownHostException ex = new UnknownHostException(generateNativeIOExceptionMessage(cause));
		ex.initCause(cause);
		return ex;
	}

	public static RuntimeException generateNativeRuntimeException(NativeIOException cause) {
		RuntimeException ex = new RuntimeException(generateNativeIOExceptionMessage(cause));
		ex.initCause(cause);
		return ex;
	}

	public static IOException generateNativeIOException(NativeIOException cause) {
		int errCode = cause.getErrorCode();
		String exceptionMsg = generateNativeIOExceptionMessage(cause);
		if ((errCode == NetErrors.J_EACCES) || (errCode == NetErrors.J_ENETDOWN) || (errCode == NetErrors.J_ENETUNREACH)
				|| (errCode == NetErrors.J_EINVAL) || (errCode == NetErrors.J_EMSGSIZE)
				|| (errCode == NetErrors.J_ENOMEM) || (errCode == NetErrors.J_EUNKNOWN)) {
			return new IOException(exceptionMsg, cause);
		}

		if (errCode == NetErrors.J_ETIMEDOUT) {
			SocketTimeoutException ex = new SocketTimeoutException(exceptionMsg);
			ex.initCause(cause);
			return ex;
		}
		if (errCode == NetErrors.J_EHOSTUNKNOWN) {
			UnknownHostException ex = new UnknownHostException(exceptionMsg);
			ex.initCause(cause);
			return ex;
		}

		if (errCode == NetErrors.J_NET_NATIVE_NOT_IMPLEMENTED) {
			throw new RuntimeException(exceptionMsg, cause);
		}

		SocketException ex = new SocketException(exceptionMsg);
		ex.initCause(cause);
		return ex;
	}

	private static String generateNativeIOExceptionMessage(NativeIOException cause) {
		int errCode = cause.getErrorCode();
		return Message.at(NetErrors.getInstance(), errCode);
	}

}
