/*
 * Copyright 2017-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.digest;

import ej.bon.ByteArray;
import ej.sni.NativeException;
import ej.sni.NativeResource;

/**
 *
 */
public class NativeMessageDigestSpi extends AbstractMessageDigestSpi {

	/**
	 * @see #nativeGetAlgorithmDescription(byte[], byte[])
	 */
	public static final int ALGORITHM_DESCRIPTION_SIZE = ByteArray.INT_SIZE * 1;

	private final int nativeAlgorithmId;

	/**
	 * Native identifier for the cipher. Initialized during engine reset.
	 */
	private int nativeId;

	/**
	 * @param nativeAlgorithmId
	 */
	public NativeMessageDigestSpi(int nativeAlgorithmId, int digestLength) {
		super(digestLength);
		this.nativeAlgorithmId = nativeAlgorithmId;
		this.nativeId = -1;
		implReset();
	}

	@Override
	void implUpdate(byte[] b, int ofs, int len) {
		nativeUpdate(this.nativeAlgorithmId, this.nativeId, b, ofs, len);
	}

	@Override
	void implDigest(byte[] out, int ofs, int len) {
		nativeDigest(this.nativeAlgorithmId, this.nativeId, out, ofs, len);
	}

	@Override
	void implReset() {

		if (this.nativeId != -1) {
			// close and reset the existing resource
			NativeResource.clearCloseOnGC(this.nativeId, nativeGetCloseId(this.nativeAlgorithmId));
			nativeClose(this.nativeAlgorithmId, this.nativeId);
			this.nativeId = -1;
		}

		int nativeAlgorithmId = this.nativeAlgorithmId;

		try {
			this.nativeId = nativeInit(nativeAlgorithmId);
			NativeResource.closeOnGC(this.nativeId, nativeGetCloseId(nativeAlgorithmId), this);

		} catch (NativeException e) {
			// An error occurs during initialization, maybe there is no space left to
			// allocate new native resource.
			// Trigger a new fresh GC and release new potential useless resources and try
			// the initialization again!
			System.gc();
			this.nativeId = nativeInit(nativeAlgorithmId);
			NativeResource.closeOnGC(this.nativeId, nativeGetCloseId(nativeAlgorithmId), this);
		}
	}

	/**
	 * Gets for the given algorithm the message digest description.
	 * <p>
	 * <code>description</code> must be filled-in with:
	 * <ul>
	 * <li>[0-3]: digestLength: length of the message digest in bytes</li>
	 * </ul>
	 *
	 * @param algorithm
	 *            Null terminated string that describes the algorithm
	 * @return algorithm ID on success or -1 on error.
	 */
	native public static int nativeGetAlgorithmDescription(byte[] algorithm, byte[] description);

	/**
	 * Returns a nativeId.
	 *
	 * @throws {@link
	 *             NativeException} on error
	 */
	native private static int nativeInit(int nativeAlgorithmId) throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link
	 *             NativeException} on error
	 */
	native public static void nativeClose(int nativeAlgorithmId, int nativeId) throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link
	 *             NativeException} on error
	 */
	native private static void nativeUpdate(int nativeAlgorithmId, int nativeId, byte[] buffer, int offset, int length)
			throws NativeException;

	/**
	 * @param nativeAlgorithmId
	 * @param nativeId
	 *
	 * @throws {@link
	 *             NativeException} on error
	 */
	native private static void nativeDigest(int nativeAlgorithmId, int nativeId, byte[] out, int offset, int length)
			throws NativeException;

	/**
	 * Gets the id of the native close function.
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 */
	native private static int nativeGetCloseId(int nativeAlgorithmId) throws NativeException;
}
