/*
 * Copyright 2018-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.keyfactory;

import java.security.KeyFactorySpi;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.spec.EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;

import com.is2t.support.security.NativePrivateKey;
import com.is2t.support.security.NativePublicKey;

import ej.sni.NativeException;
import ej.sni.NativeResource;
import ej.sni.SNI;

/**
 *
 */
public class NativeKeyFactorySpi extends KeyFactorySpi {
	private final String algorithm;
	private final int nativeAlgorithmId;

	/**
	 * SPI constructor.
	 */
	public NativeKeyFactorySpi(int nativeAlgorithmId, String algorithm) {
		this.nativeAlgorithmId = nativeAlgorithmId;
		this.algorithm = algorithm;
	}

	@Override
	protected PrivateKey engineGeneratePrivate(KeySpec keySpec) throws InvalidKeySpecException {
		EncodedKeySpec encodedKeySpec = (EncodedKeySpec) keySpec;
		String format = encodedKeySpec.getFormat();
		byte[] encodedKey = encodedKeySpec.getEncoded();

		byte[] cFormat = new byte[format.length() + 1]; // add 1 for the '\0'
		SNI.toCString(format, cFormat);

		int nativeId = nativeGetPrivateKeyData(this.nativeAlgorithmId, cFormat, encodedKey, encodedKey.length);
		PrivateKey privateKey = new NativePrivateKey(nativeId, this.algorithm);
		NativeResource.closeOnGC(nativeId, nativeGetPrivateKeyCloseId(this.nativeAlgorithmId), privateKey);
		return privateKey;
	}

	@Override
	protected PublicKey engineGeneratePublic(KeySpec keySpec) {
		EncodedKeySpec encodedKeySpec = (EncodedKeySpec) keySpec;
		String format = encodedKeySpec.getFormat();
		byte[] encodedKey = encodedKeySpec.getEncoded();

		byte[] cFormat = new byte[format.length() + 1]; // add 1 for the '\0'
		SNI.toCString(format, cFormat);

		int nativeId = nativeGetPublicKeyData(this.nativeAlgorithmId, cFormat, encodedKey, encodedKey.length);
		PublicKey publicKey = new NativePublicKey(nativeId, this.algorithm);
		NativeResource.closeOnGC(nativeId, nativeGetPublicKeyCloseId(this.nativeAlgorithmId), publicKey);
		return publicKey;
	}

	/**
	 *
	 * @param algorithm
	 *            Null terminated string that describes the algorithm
	 * @return algorithm ID on success or -1 on error.
	 */
	public static native int nativeGetAlgorithmDescription(byte[] algorithm);

	/**
	 * Generate a private key from the encoded key format.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 * @param format
	 *            the format of the key, a null terminated '\0' byte array representation of the format in String
	 * @param encodedKey
	 *            the encoded key data
	 * @param encodedKeyLength
	 *            the length of encodedKey
	 * @return the pointer of the C structure holding the key data
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetPrivateKeyData(int nativeAlgorithmId, byte[] format, byte[] encodedKey,
			int encodedKeyLength);

	/**
	 * Generate a public key from the encoded key format.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 * @param format
	 *            the format of the key, a null terminated '\0' byte array representation of the format in String
	 * @param encodedKey
	 *            the encoded key data
	 * @param encodedKeyLength
	 *            the length of encodedKey
	 * @return the pointer of the C structure holding the key data
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetPublicKeyData(int nativeAlgorithmId, byte[] format, byte[] encodedKey,
			int encodedKeyLength);

	/**
	 * Gets the id of the native close function.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetPrivateKeyCloseId(int nativeAlgorithmId);

	/**
	 * Gets the id of the native close function.
	 *
	 * @param nativeAlgorithmId
	 *            the algorithm ID
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 *             on error
	 */
	private static native int nativeGetPublicKeyCloseId(int nativeAlgorithmId);

}
