/*
 * Copyright 2018-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package com.is2t.support.security.securerandom;

import java.security.ProviderException;
import java.security.SecureRandomSpi;

import ej.sni.NativeException;
import ej.sni.NativeResource;

/**
 *
 */
public class NativeSecureRandomSpi extends SecureRandomSpi {

	/**
	 * Native identifier for the secure random engine. Initialized during engine init.
	 */
	private final int nativeId;

	/**
	 *
	 */
	public NativeSecureRandomSpi() {
		super();

		int nativeId;
		try {
			nativeId = nativeInit();
			NativeResource.closeOnGC(nativeId, nativeGetCloseId(), this);

		} catch (NativeException e) {
			// An error occurs during initialization, maybe there is no space left to
			// allocate new native resource.
			// Trigger a new fresh GC and release new potential useless resources and try
			// the initialization again!
			System.gc();
			try {
				nativeId = nativeInit();
			} catch (NativeException e2) {
				// Wrap any native exception into a ProviderException.
				throw new ProviderException(e2);
			}
		}

		// the initialization succeeds; create the associated resource.
		this.nativeId = nativeId;
	}

	@Override
	protected void engineSetSeed(byte[] seed) {
		nativeSetSeed(this.nativeId, seed, seed.length);
	}

	@Override
	protected void engineNextBytes(byte[] bytes) {
		nativeNextBytes(this.nativeId, bytes, bytes.length);
	}

	@Override
	protected byte[] engineGenerateSeed(int numBytes) {
		byte[] seed = new byte[numBytes];
		nativeGenerateSeed(this.nativeId, seed, numBytes);
		return seed;
	}

	/**
	 * Initializes the native resource.
	 *
	 * @return a nativeId
	 * @throws NativeException
	 */
	native public static int nativeInit() throws NativeException;

	/**
	 * Sets the PRNG seed.
	 *
	 * @param nativeId
	 *            the resource id
	 * @param seed
	 *            the seed to use
	 * @throws NativeException
	 */
	native public static void nativeSetSeed(int nativeId, byte[] seed, int seedLength) throws NativeException;

	/**
	 * Generates random bytes.
	 *
	 * @param nativeId
	 *            the resource id
	 * @param bytes
	 *            the array to fill with random bytes
	 * @param numBytes
	 *            the number of bytes to generate
	 * @throws NativeException
	 */
	native public static void nativeNextBytes(int nativeId, byte[] bytes, int numBytes) throws NativeException;

	/**
	 * Generates a seed.
	 *
	 * @param nativeId
	 *            the resource id
	 * @param bytes
	 *            the seed array to fill with random bytes
	 * @param numBytes
	 *            the seed size in bytes
	 * @throws NativeException
	 */
	native public static void nativeGenerateSeed(int nativeId, byte[] bytes, int numBytes) throws NativeException;

	/**
	 * Closes the native resource.
	 *
	 * @param nativeId
	 *            the resource id
	 * @throws NativeException
	 */
	native public static void nativeClose(int nativeId) throws NativeException;

	/**
	 * Gets the id of the native close function.
	 *
	 * @return the id of the static native close function
	 * @throws NativeException
	 */
	native private static int nativeGetCloseId() throws NativeException;
}
