/**
 * Copyright 2024 NXP
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
package com.nxp.io.gpio;

import com.nxp.io.gpio.natives.GPIONatives;

@SuppressWarnings({ "javadoc", "nls" })
public class GPIO {

	private final int pin;
	private final int direction;
	private final int pullConfig;

	public static final int GPIO_DIGITAL_INPUT = 0;
	public static final int GPIO_DIGITAL_OUTPUT = 1;
	public static final int GPIO_NO_PULL = 0;
	public static final int GPIO_PULL_UP = 1;
	public static final int GPIO_PULL_DOWN = 2;
	private static final int kPortError = -1;
	private static final int kPinError = -2;

	public GPIO(int pin, int direction, int pullConfig) {
		if (pin == -1) {
			throw new IllegalArgumentException(
					"This pin number is not valid ! Please check your pin has been configured correctly inside gpio.properties file");
		}
		if (direction < GPIO_DIGITAL_INPUT || direction > GPIO_DIGITAL_OUTPUT) {
			throw new IllegalArgumentException("Direction should be GPIO_DIGITAL_INPUT or GPIO_DIGITAL_OUTPUT");
		}
		if (pullConfig < GPIO_NO_PULL || pullConfig > GPIO_PULL_DOWN) {
			throw new IllegalArgumentException("pullConfig should be GPIO_NO_PULL, GPIO_PULL_UP or GPIO_PULL_DOWN");
		}
		this.pin = pin;
		this.direction = direction;
		this.pullConfig = pullConfig;
		int status = GPIONatives.nativeInitGPIOPin(this.pin, this.direction, this.pullConfig);
		checkStatus(status);
	}

	private static void checkStatus(int status) {
		if (status == kPortError) {
			throw new IllegalArgumentException("GPIO Port Error");
		} else if (status == kPinError) {
			throw new IllegalArgumentException("GPIO Pin Error");
		}
	}

	public int getPin() {
		return this.pin;
	}

	public int getDirection() {
		return this.direction;
	}

	public int getPullConfig() {
		return this.pullConfig;
	}

	public int readPin() {
		int status = GPIONatives.nativeReadGPIOPin(this.pin);
		checkStatus(status);
		return status;
	}

	public int writePin(int level) {
		if (level < 0 || level > 1) {
			throw new IllegalArgumentException("Level value should be 0 or 1");
		}
		int status = GPIONatives.nativeWriteGPIOPin(this.pin, level);
		checkStatus(status);
		return status;
	}

	public int togglePin() {
		int status = GPIONatives.nativeToggleGPIOPin(this.pin);
		checkStatus(status);
		return status;
	}

}
