/*
 * Copyright (c) 1995, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright 2015-2022 MicroEJ Corp. This file has been modified and/or created by MicroEJ Corp.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation. The Copyright holder(s) designates
 * this particular file as subject to the "Classpath" exception as provided
 * by the Copyright holder(s) in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact the Copyright holder(s) if you need additional information
 * or have any questions.
 * (if Oracle: 500 Oracle Parkway, Redwood Shores, CA 94065 USA or visit www.oracle.com)
 */

package java.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import sun.net.ConnectionResetException;

/**
 * Default Socket Implementation. This implementation does not implement any security checks. Note this class should
 * <b>NOT</b> be public.
 *
 * @author Steven B. Byrne
 */
public abstract class AbstractPlainSocketImpl extends SocketImpl {
	/* instance variable for SO_TIMEOUT */
	int timeout; // timeout in millisec
	// traffic class
	private int trafficClass;

	private SocketInputStream socketInputStream = null;
	private SocketOutputStream socketOutputStream = null;

	/* number of threads using the FileDescriptor */
	protected int fdUseCount = 0;

	/* lock when increment/decrementing fdUseCount */
	protected final Object fdLock = new Object();

	/* indicates a close is pending on the file descriptor */
	protected boolean closePending = false;

	/* indicates connection reset state */
	private final int CONNECTION_NOT_RESET = 0;
	private final int CONNECTION_RESET_PENDING = 1;
	private final int CONNECTION_RESET = 2;
	private int resetState;
	private final Object resetLock = new Object();

	/*
	 * whether this Socket is a stream (TCP) socket or not (UDP)
	 */
	protected boolean stream;

	private boolean isClosed;

	/**
	 * Creates a socket with a boolean that specifies whether this is a stream socket (true) or an unconnected UDP
	 * socket (false).
	 */
	@Override
	protected synchronized void create(boolean stream) throws IOException {
		this.stream = stream;
		socketCreate(stream);
		if (this.socket != null) {
			this.socket.setCreated();
		}
	}

	/**
	 * IS2T additional method<br>
	 * Creates a server socket
	 */
	@Override
	protected synchronized void createServer() throws IOException {
		serverSocketCreate();
		if (this.serverSocket != null) {
			this.serverSocket.setCreated();
		}
	}

	/**
	 * Creates a socket and connects it to the specified port on the specified host.
	 *
	 * @param host
	 *            the specified host
	 * @param port
	 *            the specified port
	 */
	@Override
	protected void connect(String host, int port) throws UnknownHostException, IOException {
		boolean connected = false;
		try {
			InetAddress address = InetAddress.getByName(host);
			this.port = port;
			this.address = address;

			connectToAddress(address, port, this.timeout);
			connected = true;
		} finally {
			if (!connected) {
				try {
					close();
				} catch (IOException ioe) {
					/*
					 * Do nothing. If connect threw an exception then it will be passed up the call stack
					 */
				}
			}
		}
	}

	/**
	 * Creates a socket and connects it to the specified address on the specified port.
	 *
	 * @param address
	 *            the address
	 * @param port
	 *            the specified port
	 */
	@Override
	protected void connect(InetAddress address, int port) throws IOException {
		this.port = port;
		this.address = address;

		try {
			connectToAddress(address, port, this.timeout);
			return;
		} catch (IOException e) {
			// everything failed
			close();
			throw e;
		}
	}

	/**
	 * Creates a socket and connects it to the specified address on the specified port.
	 *
	 * @param address
	 *            the address
	 * @param timeout
	 *            the timeout value in milliseconds, or zero for no timeout.
	 * @throws IOException
	 *             if connection fails
	 * @throws IllegalArgumentException
	 *             if address is null or is a SocketAddress subclass not supported by this socket
	 * @since 1.4
	 */
	@Override
	protected void connect(SocketAddress address, int timeout) throws IOException {
		boolean connected = false;
		try {
			if (address == null || !(address instanceof InetSocketAddress)) {
				throw new IllegalArgumentException("unsupported address type");
			}
			InetSocketAddress addr = (InetSocketAddress) address;
			if (addr.isUnresolved()) {
				throw new UnknownHostException(addr.getHostName());
			}
			this.port = addr.getPort();
			this.address = addr.getAddress();

			connectToAddress(this.address, this.port, timeout);
			connected = true;
		} finally {
			if (!connected) {
				try {
					close();
				} catch (IOException ioe) {
					/*
					 * Do nothing. If connect threw an exception then it will be passed up the call stack
					 */
				}
			}
		}
	}

	private void connectToAddress(InetAddress address, int port, int timeout) throws IOException {
		if (address.isAnyLocalAddress()) {
			doConnect(InetAddress.getLocalHost(), port, timeout);
		} else {
			doConnect(address, port, timeout);
		}
	}

	@Override
	public void setOption(int opt, Object val) throws SocketException {
		if (isClosedOrPending()) {
			throw new SocketException("Socket Closed");
		}
		boolean on = true;
		switch (opt) {
		/*
		 * check type safety b4 going native. These should never fail, since only java.Socket* has access to
		 * PlainSocketImpl.setOption().
		 */
		case SO_LINGER:
			if (val == null || (!(val instanceof Integer) && !(val instanceof Boolean))) {
				throw new SocketException("Bad parameter for option");
			}
			if (val instanceof Boolean) {
				/* true only if disabling - enabling should be Integer */
				on = false;
			}
			break;
		case SO_TIMEOUT:
			if (val == null || (!(val instanceof Integer))) {
				throw new SocketException("Bad parameter for SO_TIMEOUT");
			}
			int tmp = ((Integer) val).intValue();
			if (tmp < 0) {
				throw new IllegalArgumentException("timeout < 0");
			}
			this.timeout = tmp;
			return;
		case IP_TOS:
			if (val == null || !(val instanceof Integer)) {
				throw new SocketException("bad argument for IP_TOS");
			}
			this.trafficClass = ((Integer) val).intValue();
			break;
		case SO_BINDADDR:
			throw new SocketException("Cannot re-bind socket");
		case TCP_NODELAY:
			if (val == null || !(val instanceof Boolean)) {
				throw new SocketException("bad parameter for TCP_NODELAY");
			}
			on = ((Boolean) val).booleanValue();
			break;
		case SO_SNDBUF:
		case SO_RCVBUF:
			if (val == null || !(val instanceof Integer) || !(((Integer) val).intValue() > 0)) {
				throw new SocketException("bad parameter for SO_SNDBUF " + "or SO_RCVBUF");
			}
			break;
		case SO_KEEPALIVE:
			if (val == null || !(val instanceof Boolean)) {
				throw new SocketException("bad parameter for SO_KEEPALIVE");
			}
			on = ((Boolean) val).booleanValue();
			break;
		case SO_OOBINLINE:
			if (val == null || !(val instanceof Boolean)) {
				throw new SocketException("bad parameter for SO_OOBINLINE");
			}
			on = ((Boolean) val).booleanValue();
			break;
		case SO_REUSEADDR:
			if (val == null || !(val instanceof Boolean)) {
				throw new SocketException("bad parameter for SO_REUSEADDR");
			}
			on = ((Boolean) val).booleanValue();
			break;
		default:
			throw new SocketException("unrecognized TCP option: " + opt);
		}

		socketSetOption(opt, on, val);
	}

	@Override
	public Object getOption(int opt) throws SocketException {
		if (isClosedOrPending()) {
			throw new SocketException("Socket Closed");
		}
		Object ret;

		switch (opt) {

		case SO_TIMEOUT:
			ret = new Integer(this.timeout);
			break;
		case IP_TOS:
			ret = socketGetOption(opt);
			if (((Integer) ret).intValue() == -1) // ipv6 tos
			{
				ret = new Integer(this.trafficClass);
			}
			break;
		case SO_KEEPALIVE:
		case TCP_NODELAY:
		case SO_OOBINLINE:
		case SO_LINGER:
		case SO_REUSEADDR:
		case SO_BINDADDR:
		case SO_SNDBUF:
		case SO_RCVBUF:
			ret = socketGetOption(opt);
			break;
		default:
			throw new SocketException("invalid option: " + opt);
		}
		return ret;
	}

	/**
	 * The workhorse of the connection operation. Tries several times to establish a connection to the given <host,
	 * port>. If unsuccessful, throws an IOException indicating what went wrong.
	 */

	synchronized void doConnect(InetAddress address, int port, int timeout) throws IOException {
		try {
			acquireFD();
			try {
				socketConnect(address, port, timeout);
				/* socket may have been closed during poll/select */
				synchronized (this.fdLock) {
					if (this.closePending) {
						throw new SocketException("Socket closed");
					}
				}
				// If we have a ref. to the Socket, then sets the flags
				// created, bound & connected to true.
				// This is normally done in Socket.connect() but some
				// subclasses of Socket may call impl.connect() directly!
				if (this.socket != null) {
					this.socket.setBound();
					this.socket.setConnected();
				}
			} finally {
				releaseFD();
			}
		} catch (IOException e) {
			close();
			throw e;
		}
	}

	/**
	 * Binds the socket to the specified address of the specified local port.
	 *
	 * @param address
	 *            the address
	 * @param lport
	 *            the port
	 */
	@Override
	protected synchronized void bind(InetAddress address, int lport) throws IOException {
		socketBind(address, lport);
		if (this.socket != null) {
			this.socket.setBound();
		}
		if (this.serverSocket != null) {
			this.serverSocket.setBound();
		}
	}

	/**
	 * Listens, for a specified amount of time, for connections.
	 *
	 * @param count
	 *            the amount of time to listen for connections
	 */
	@Override
	protected synchronized void listen(int count) throws IOException {
		socketListen(count);
	}

	/**
	 * Accepts connections.
	 *
	 * @param s
	 *            the connection
	 */
	@Override
	protected void accept(SocketImpl s) throws IOException {
		acquireFD();
		try {
			socketAccept(s);
		} finally {
			releaseFD();
		}
	}

	/**
	 * Gets an InputStream for this socket.
	 */
	@Override
	protected synchronized InputStream getInputStream() throws IOException {
		synchronized (this.fdLock) {
			if (isClosedOrPending()) {
				throw new IOException("Socket Closed");
			}
			if (this.socketInputStream == null) {
				this.socketInputStream = new SocketInputStream(this);
			}
		}
		return this.socketInputStream;
	}

	void setInputStream(SocketInputStream in) {
		this.socketInputStream = in;
	}

	/**
	 * Gets an OutputStream for this socket.
	 */
	@Override
	protected synchronized OutputStream getOutputStream() throws IOException {
		synchronized (this.fdLock) {
			if (isClosedOrPending()) {
				throw new IOException("Socket Closed");
			}
			if (this.socketOutputStream == null) {
				this.socketOutputStream = new SocketOutputStream(this);
			}
		}
		return this.socketOutputStream;
	}

	void setAddress(InetAddress address) {
		this.address = address;
	}

	void setPort(int port) {
		this.port = port;
	}

	void setLocalPort(int localport) {
		this.localport = localport;
	}

	/**
	 * Returns the number of bytes that can be read without blocking.
	 */
	@Override
	protected synchronized int available() throws IOException {
		if (isClosedOrPending()) {
			throw new IOException("Stream closed.");
		}

		/*
		 * If connection has been reset or shut down for input, then return 0 to indicate there are no buffered bytes.
		 */
		if (isConnectionReset()) {
			return 0;
		}

		/*
		 * If no bytes available and we were previously notified of a connection reset then we move to the reset state.
		 *
		 * If are notified of a connection reset then check again if there are bytes buffered on the socket.
		 */
		int n = 0;
		try {
			n = socketAvailable();
			if (n == 0 && isConnectionResetPending()) {
				setConnectionReset();
			}
		} catch (ConnectionResetException exc1) {
			setConnectionResetPending();
			try {
				n = socketAvailable();
				if (n == 0) {
					setConnectionReset();
				}
			} catch (ConnectionResetException exc2) {
			}
		}
		return n;
	}

	/**
	 * Closes the socket.
	 */
	@Override
	protected void close() throws IOException {
		synchronized (this.fdLock) {
			if (!this.isClosed) {
				if (this.fdUseCount == 0) {
					if (this.closePending) {
						return;
					}
					this.closePending = true;
					/*
					 * We close the FileDescriptor in two-steps - first the "pre-close" which closes the socket but
					 * doesn't release the underlying file descriptor. This operation may be lengthy due to
					 * untransmitted data and a long linger interval. Once the pre-close is done we do the actual socket
					 * to release the fd.
					 */
					try {
						socketPreClose();
					} finally {
						socketClose();
					}
					this.isClosed = false;
					return;
				} else {
					/*
					 * If a thread has acquired the fd and a close isn't pending then use a deferred close. Also
					 * decrement fdUseCount to signal the last thread that releases the fd to close it.
					 */
					if (!this.closePending) {
						this.closePending = true;
						this.fdUseCount--;
						socketPreClose();
					}
				}
			}
		}
	}

	@Override
	void reset() throws IOException {
		socketClose();
		if (!this.isClosed) {
			socketClose();
		}
		this.isClosed = true;
		super.reset();
	}

	/**
	 * Cleans up if the user forgets to close it.
	 */
	protected void finalize() throws IOException {
		close();
	}

	/*
	 * "Acquires" and returns the FileDescriptor for this impl
	 *
	 * A corresponding releaseFD is required to "release" the FileDescriptor.
	 */
	/* CLDC-1.1-COMPLIANCE FileDescriptor */ void acquireFD() {
		synchronized (this.fdLock) {
			this.fdUseCount++;
		}
	}

	/*
	 * "Release" the FileDescriptor for this impl.
	 *
	 * If the use count goes to -1 then the socket is closed.
	 */
	void releaseFD() {
		synchronized (this.fdLock) {
			this.fdUseCount--;
			if (this.fdUseCount == -1) {
				try {
					socketClose();
				} catch (IOException e) {
				}
				if (!this.isClosed) {
					try {
						socketClose();
					} catch (IOException e) {
					} finally {
						this.isClosed = true;
					}
				}
			}
		}
	}

	public boolean isConnectionReset() {
		synchronized (this.resetLock) {
			return (this.resetState == this.CONNECTION_RESET);
		}
	}

	public boolean isConnectionResetPending() {
		synchronized (this.resetLock) {
			return (this.resetState == this.CONNECTION_RESET_PENDING);
		}
	}

	public void setConnectionReset() {
		synchronized (this.resetLock) {
			this.resetState = this.CONNECTION_RESET;
		}
	}

	public void setConnectionResetPending() {
		synchronized (this.resetLock) {
			if (this.resetState == this.CONNECTION_NOT_RESET) {
				this.resetState = this.CONNECTION_RESET_PENDING;
			}
		}

	}

	/*
	 * Return true if already closed or close is pending
	 */
	public boolean isClosedOrPending() {
		/*
		 * Lock on fdLock to ensure that we wait if a close is in progress.
		 */
		synchronized (this.fdLock) {
			if (this.closePending || (this.isClosed)) {
				return true;
			} else {
				return false;
			}
		}
	}

	/*
	 * Return the current value of SO_TIMEOUT
	 */
	public int getTimeout() {
		return this.timeout;
	}

	/*
	 * "Pre-close" a socket by dup'ing the file descriptor - this enables the socket to be closed without releasing the
	 * file descriptor.
	 */
	private void socketPreClose() throws IOException {
		socketClose0(true);
	}

	/*
	 * Close the socket (and release the file descriptor).
	 */
	protected void socketClose() throws IOException {
		socketClose0(false);
	}

	protected abstract void socketCreate(boolean stream) throws IOException;

	protected abstract void socketConnect(InetAddress address, int port, int timeout) throws IOException;

	protected abstract void socketBind(InetAddress address, int port) throws IOException;

	protected abstract void socketListen(int count) throws IOException;

	protected abstract void socketAccept(SocketImpl s) throws IOException;

	protected abstract int socketAvailable() throws IOException;

	protected abstract void socketClose0(boolean useDeferredClose) throws IOException;

	protected abstract void socketSetOption(int cmd, boolean on, Object value) throws SocketException;

	protected abstract Object socketGetOption(int opt) throws SocketException;

	// IS2T additional methods
	protected abstract void serverSocketCreate() throws IOException;

	protected abstract int socketRead(byte b[], int off, int len) throws IOException;

	protected abstract void socketWrite(byte b[], int off, int len) throws IOException;
}
